#include <torch/extension.h>
#include <ATen/ATen.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <cmath>

// Macros for input checking
#define CHECK_CUDA(x) TORCH_CHECK(x.is_cuda(), #x " must be a CUDA tensor")
#define CHECK_CONTIGUOUS(x) TORCH_CHECK(x.is_contiguous(), #x " must be contiguous")
#define CHECK_INPUT(x) \
    CHECK_CUDA(x);     \
    CHECK_CONTIGUOUS(x)


// Fused activation kernel using a grid-stride loop for flexibility and efficient load balancing
template <typename scalar_t>
__global__ void fused_activation_kernel(
    const scalar_t* __restrict__ x_in,
    scalar_t* __restrict__ x_out,
    const size_t size) {

    size_t idx = blockIdx.x * blockDim.x + threadIdx.x;
    const scalar_t one = static_cast<scalar_t>(1);
    const scalar_t two = static_cast<scalar_t>(2);
    
    // Grid-stride loop handles arbitrary tensor sizes
    for (; idx < size; idx += blockDim.x * gridDim.x) {
        scalar_t x = x_in[idx];

        // Swish activation: x * sigmoid(x)
        scalar_t sigmoid_x = one / (one + exp(-x));
        x = x * sigmoid_x;
        
        // Scale down value by 2
        x = x / two;
        
        // Clamp between -1 and 1 (first clamp for numerical stability)
        x = max(min(x, one), -one);

        // Tanh activation
        x = tanh(x);

        // Second clamping for extra safety (tanh already returns in [-1, 1] but to enforce strict bounds)
        x = max(min(x, one), -one);
        
        x_out[idx] = x;
    }
}

// CUDA forward function fusing linear transformation with activation
// Note: We still use torch::addmm for the linear part since cuBLAS GEMM is highly optimized
// and then fuse the activation in one efficient kernel.

torch::Tensor module_forward_cuda(
    torch::Tensor x,
    torch::Tensor weight,
    torch::Tensor bias) {

    // Execute linear transformation: x_linear = bias + x * weight^T
    auto x_linear = torch::addmm(bias, x, weight.t());
    auto x_out = torch::empty_like(x_linear);

    // Compute total number of elements
    size_t numel = x_linear.numel();

    // Use a 1D grid-stride loop kernel for the activation part
    // Choosing a block size of 256 threads to balance occupancy
    const int threads = 256;
    const int blocks = (numel + threads - 1) / threads;

    AT_DISPATCH_FLOATING_TYPES_AND_HALF(x_linear.scalar_type(), "module_forward_cuda", ([&] {
        fused_activation_kernel<scalar_t><<<blocks, threads>>>(
            x_linear.data_ptr<scalar_t>(),
            x_out.data_ptr<scalar_t>(),
            numel);
    }));

    return x_out;
}

// C++ interface that wraps the CUDA implementation
torch::Tensor module_forward(
    torch::Tensor x,
    torch::Tensor weight,
    torch::Tensor bias) {
    CHECK_INPUT(x);
    CHECK_INPUT(weight);
    CHECK_INPUT(bias);
    return module_forward_cuda(x, weight, bias);
}

// Binding the forward function to Python via PyBind11
PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &module_forward, "Fused linear and activation forward function (CUDA)");
}
